@@ -5,8 +5,13 @@ use crate::{TextDuplexer, TextReader, TextSubstr};
 use basic_text_internals::{
     is_basic_text_end, is_basic_text_start, replace,
     unicode::{
-        BEL, BOM, CAN, CGJ, DEL, ESC, LS, MAX_UTF8_SIZE, NEL, NORMALIZATION_BUFFER_SIZE, PS, REPL,
+        BEL, BOM, CAN, CGJ, DEL, ESC, LS, MAX_UTF8_SIZE, NEL, NORMALIZATION_BUFFER_SIZE, PS,
     },
+    unicode_normalization::{
+        char::is_public_assigned, is_nfc_stream_safe_quick, IsNormalized, Recompositions,
+        Replacements, StreamSafe, UnicodeNormalization,
+    },
+    IsolateUnassigned, PreNormalization,
 };
 use layered_io::{default_read, HalfDuplexLayered, Status, WriteLayered};
 use std::{
@@ -16,10 +21,6 @@ use std::{
     mem::take,
     str,
 };
-use unicode_normalization::{
-    is_nfc_stream_safe_quick, IsNormalized, Recompositions, Replacements, StreamSafe,
-    UnicodeNormalization,
-};
 use utf8_io::{ReadStrLayered, WriteStr};
 
 /// Abstract over `TextReader` and the reader half of `TextDuplexer`.
@@ -79,7 +80,8 @@ pub(crate) struct TextInput {
     queue: VecDeque<char>,
 
     /// An iterator over the chars in `self.queue`.
-    ssnfc_iter: Recompositions<StreamSafe<Replacements<vec_deque::IntoIter<char>>>>,
+    ssnfc_iter:
+        Recompositions<StreamSafe<Replacements<IsolateUnassigned<vec_deque::IntoIter<char>>>>>,
 
     /// The number of characters in the queue which are already verified to be
     /// Stream-Safe NFC and can skip normalization.
@@ -116,6 +118,7 @@ impl TextInput {
             queue,
             ssnfc_iter: VecDeque::<char>::new()
                 .into_iter()
+                .isolate_unassigned()
                 .cjk_compat_variants()
                 .stream_safe()
                 .nfc(),
@@ -252,12 +255,23 @@ impl TextInput {
                     let index = self.queue.len() - last_boundary;
                     if is_nfc_stream_safe_quick(self.queue.iter().take(index).copied())
                         == IsNormalized::Yes
+                        && self
+                            .queue
+                            .iter()
+                            .take(index)
+                            .copied()
+                            .all(is_public_assigned)
                     {
                         self.quick = index - 1;
                         self.queue.pop_front()
                     } else {
                         let tmp = self.queue.drain(..index).collect::<VecDeque<char>>();
-                        self.ssnfc_iter = tmp.into_iter().cjk_compat_variants().stream_safe().nfc();
+                        self.ssnfc_iter = tmp
+                            .into_iter()
+                            .isolate_unassigned()
+                            .cjk_compat_variants()
+                            .stream_safe()
+                            .nfc();
                         self.ssnfc_iter.next()
                     }
                 }
@@ -283,6 +297,7 @@ impl TextInput {
                             self.state = State::Ground(true);
                         }
                         '\r' => self.state = State::Cr,
+                        '\x0c' => self.state = State::Ff,
                         ESC => self.state = State::Esc,
                         mut c => {
                             self.state = State::Ground(false);
@@ -318,6 +333,17 @@ impl TextInput {
                         }
                     }
 
+                    (State::Ff, c) => {
+                        if c != '\x0c' {
+                            if c != '\n' && c != '\r' {
+                                self.queue.push_back(' ');
+                            }
+                            self.expect_starter = false;
+                            self.state = State::Ground(false);
+                            continue;
+                        }
+                    }
+
                     (State::Esc, '[') => self.state = State::CsiStart,
                     (State::Esc, ']') => self.state = State::Osc,
                     (State::Esc, ESC) => self.state = State::Esc,
@@ -325,8 +351,6 @@ impl TextInput {
                         self.state = State::Ground(false);
                     }
                     (State::Esc, _) => {
-                        self.queue.push_back(REPL);
-                        self.expect_starter = false;
                         self.state = State::Ground(false);
                         continue;
                     }
@@ -411,12 +435,12 @@ impl TextInput {
                     internals.impl_().expect_starter = false;
                     internals.impl_().state = State::Ground(true);
                 }
-                State::Esc => {
-                    internals.impl_().queue.push_back(REPL);
+                State::Ff => {
+                    internals.impl_().queue.push_back(' ');
                     internals.impl_().expect_starter = false;
                     internals.impl_().state = State::Ground(false);
                 }
-                State::CsiStart | State::Csi | State::Osc | State::Linux => {
+                State::Esc | State::CsiStart | State::Csi | State::Osc | State::Linux => {
                     internals.impl_().state = State::Ground(false);
                 }
             }
@@ -559,6 +583,9 @@ enum State {
     // After a '\r'.
     Cr,
 
+    // After a '\x0c'.
+    Ff,
+
     // After a '\x1b'.
     Esc,
 
