@@ -5,7 +5,10 @@ use crate::{TextDuplexer, TextSubstr, TextWriter};
 use basic_text_internals::{
     is_basic_text_end, is_basic_text_start,
     unicode::{BOM, ESC, MAX_UTF8_SIZE, SUB},
-    BasicTextError, Categorize,
+    unicode_normalization::{
+        char::is_public_assigned, is_nfc_stream_safe_quick, IsNormalized, UnicodeNormalization,
+    },
+    BasicTextError, PreNormalization,
 };
 #[cfg(can_vector)]
 use layered_io::default_is_write_vectored;
@@ -19,7 +22,6 @@ use std::{
     rc::Rc,
     str,
 };
-use unicode_normalization::{is_nfc_stream_safe_quick, IsNormalized, UnicodeNormalization};
 use utf8_io::{ReadStrLayered, WriteStr};
 
 /// Abstract over `TextWriter` and the writer half of `TextDuplexer`.
@@ -131,14 +133,14 @@ impl TextOutput {
 
     #[inline]
     pub(crate) fn with_bom_compatibility<Inner: WriteStr + WriteLayered>(
-        internals: &mut Inner,
+        inner: &mut Inner,
     ) -> io::Result<Self> {
         let result = Self::new();
 
         let mut bom_bytes = [0_u8; MAX_UTF8_SIZE];
         let bom_len = BOM.encode_utf8(&mut bom_bytes).len();
         // Safety: `bom_bytes` is valid UTF-8 because we just encoded it.
-        internals.write_all(&bom_bytes[..bom_len])?;
+        inner.write_all(&bom_bytes[..bom_len])?;
 
         // The BOM is not part of the logical content, so leave the stream in
         // Ground(Ground::Newline) mode, meaning we don't require a newline if
@@ -297,10 +299,12 @@ impl TextOutput {
     ) -> Result<(), BasicTextError> {
         let error = Rc::new(RefCell::new(None));
 
-        if is_nfc_stream_safe_quick(s.chars()) == IsNormalized::Yes {
-            // Fast path: Data is already Stream-Safe and NFC. Just check it
-            // for errors.
-            for c in Categorize::new(s.chars(), Rc::clone(&error)) {
+        if is_nfc_stream_safe_quick(s.chars()) == IsNormalized::Yes
+            && s.chars().all(is_public_assigned)
+        {
+            // Fast path: Data is already Stream-Safe NFC and assigned. Just
+            // check for errors.
+            for c in s.chars().categorize(Rc::clone(&error)) {
                 Self::state_machine_char(internals, c, &error)?;
             }
         } else {
@@ -316,8 +320,12 @@ impl TextOutput {
         s: &str,
         error: &Rc<RefCell<Option<BasicTextError>>>,
     ) -> Result<(), BasicTextError> {
-        // Slow path: Compute Stream-Safe and NFC.
-        for c in Categorize::new(s.chars(), Rc::clone(error))
+        // Slow path: Compute Stream-Safe NFC, isolate unassigned scalar
+        // values, and check for errors.
+        for c in s
+            .chars()
+            .categorize(Rc::clone(error))
+            .isolate_unassigned()
             .cjk_compat_variants()
             .stream_safe()
             .nfc()
@@ -399,21 +407,21 @@ impl TextOutput {
                 Self::prepare_failure(internals);
                 Err(io::Error::new(
                     io::ErrorKind::InvalidData,
-                    "output text stream closed after a ZWJ or Prepend",
+                    "strict text stream ended after a ZWJ or Prepend",
                 ))
             }
             State::Ground(Ground::Other) => {
                 Self::prepare_failure(internals);
                 Err(io::Error::new(
                     io::ErrorKind::InvalidData,
-                    "output text stream must end with newline",
+                    "strict text stream must end with newline",
                 ))
             }
             State::Esc | State::Csi => {
                 Self::prepare_failure(internals);
                 Err(io::Error::new(
                     io::ErrorKind::InvalidData,
-                    "incomplete escape sequence at end of output text stream",
+                    "incomplete escape sequence at end of strict text stream",
                 ))
             }
         }
@@ -491,14 +499,14 @@ impl TextOutput {
                 Self::prepare_failure(internals);
                 return Err(io::Error::new(
                     io::ErrorKind::InvalidData,
-                    "output text stream flushed after a ZWJ or Prepend",
+                    "strict text stream flushed after a ZWJ or Prepend",
                 ));
             }
             State::Ground(_) => (),
             State::Esc | State::Csi => {
                 return Err(io::Error::new(
                     io::ErrorKind::InvalidData,
-                    "output text stream flushed while an escape sequence was in progress",
+                    "strict text stream flushed while an escape sequence was in progress",
                 ))
             }
         }
@@ -561,7 +569,7 @@ impl Drop for TextOutput {
         if let State::Ground(Ground::Newline) = self.state {
             // oll korrect
         } else {
-            panic!("output text stream not ended with newline");
+            panic!("strict text stream not ended with newline");
         }
     }
 }
@@ -569,7 +577,7 @@ impl Drop for TextOutput {
 enum Ground {
     // We just saw a '\n'.
     Newline,
-    // We just as a ZWJ or a Prepend.
+    // We just saw a ZWJ or a Prepend.
     ZwjOrPrepend,
     // Otherwise.
     Other,
